package oracle.forms.fd.jdapiutil;

import java.awt.Component;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import oracle.forms.jdapi.*;
import java.io.*;

import javax.swing.JOptionPane;


/*************************************************************** 
 *  Forms JAVA API (JDAPI): - <em>JDAPIUtil</em><br>
 *  This JavaBean allows doing modifications on Forms modules
 *  based on the content of an xml file.
 *
 * @author F. Degrelle
 * @author Oct-2007
 * @version 1.4
 ***************************************************************/

public class JDAPIUtil extends Component
{
  // constants
  public final static int OVERRIDE = 0 ;
  public final static int BEFORE   = 1 ;
  public final static int AFTER    = 2 ;
  // variables
  public static String      sModule_name    = "JDAPIUtil";
  public static String      sModule_version = "1.4";
  String      sBackupDir = "" ;  // module backup directory
  String      sTargetDir = "" ;  // module target directory
  String      sModuleName = "";   // module name
  String      sXMLfile = null ;   // XML configuration file name
  String      sBackExt = null ;   // backup extension
  String      sNewExt  = null ;   // new file extension
  String      sLibCase = null ;   // libraries case conversion  
  String      sLogFile = null ;   // log filename
  String      sConnect = "" ;     // connection string  
  String      sFileNew = "" ;     // new filename  
  String      sComment = "" ;
  String      sFileContent = "" ;
  String      sCurrentModule = "" ;
  String      sErrorCode = "^ERROR^" ;
  JdapiModule mod = null ;
  FormModule  fm = null ;  
  List        module_list  = new ArrayList(5) ; 
  boolean     bLog = true ;
  boolean     bLogOpen = false ;
  boolean     bCompile = false ;
  boolean     bFileOk  = false ;
  PrintWriter fileOut = null;
  File        f = null;
  // path
  String      sPath = null ; // file path
  String      sFile = null ; // file name
  String      sExt = "" ;    // file extension
  int         ie = 0 ;
  // generation date
  Date        today = new Date();
  SimpleDateFormat formater = new SimpleDateFormat("EEEEE dd MMMMM yyyy H:m:s");
  String      dateNewFormat = formater.format(today);   

    
  /**--------------------------------------*
   *   public procedure to call.           *
   *   the first argument : sXMLFileName   *
   *   must contain a valid XML filename   *
   *---------------------------------------*/
  public void doProcess(String sXMLFileName)
  {
    //Print header
    log(sModule_name + " Version " + sModule_version);

    try
    {

      if (sXMLFileName == null)
        throw new ErrorException("XML configuration file name must be supplied");      
      
      sXMLfile = sXMLFileName ;
      
      /**------------------------------------------*
       * open and read the xml configuration file  * 
       *-------------------------------------------*/
      log("loading XML configuration file: " + sXMLfile);
      sFileContent = readFile(sXMLfile) ;
      if(bFileOk)
      {         
        int j = 0 ;
        String st = "", st2="" ;
        
        /**------------------------------*
         *    configuration  properties  *
         *-------------------------------*/
        st = getXMLTag(sFileContent, "configuration", null) ;
        if(st != null)
        {            

          //
          // log file name
          //
          sLogFile = getXMLTag(sFileContent, "log_file_name", null) ;
          // open the log file
          if(sLogFile != null) 
          { 
            if(! openLogFile(sLogFile)) System.exit(1); 
          }
          else log("!! no log filename given !!");
        
          //Print header
          log(sModule_name + " Version " + sModule_version+"  "+dateNewFormat); 
          log(" ");
          log(" * XML filename       : "+sXMLfile);          
          if(bLogOpen) log(" * logging file       : "+sLogFile);
          
          // ----------------------------
          //  list of modules to process
          // ----------------------------
          st2 = getXMLTag(sFileContent, "module_list", null) ;
          if(st2 == null) throw new ErrorException("<module_list>: no module list given");
          List lst = getXMLTagArray(st2, "name") ;
          if(lst.size()>0)
          {
              module_list  = new ArrayList(5) ; 
              for(j=0; j<lst.size(); j++)
              {
                  String sMod = (String)lst.get(j) ;
                  File fx = new File(sMod);
                  // directory given ?
                  if(fx.isDirectory())
                  {
                    File files[] = fx.listFiles();
                    for(int k=0; k<files.length; k++)
                    {
                      sFile = files[k].getAbsolutePath();
                      // file extension
                      ie = sFile.indexOf(".") ;
                      sExt = sFile.substring(ie+1);
                      if(sExt.equalsIgnoreCase("FMB"))
                      {
                        // add module to the list
                        log(" > module to process  : "+sFile);
                        module_list.add(sFile);                        
                      }
                    }
                  }
                  else
                  {
                    // add module to the list
                    log(" > module to process  : "+sMod);
                    module_list.add(sMod);
                  }
              }
          }
          else throw new ErrorException("<module_list>: no module list given");
          
          //
          // backup directory
          //
          sBackupDir = getXMLTag(sFileContent, "backup_directory", "./") ;
          log(" * backup directory   : "+sBackupDir);

          //
          // target directory
          //
          sTargetDir = getXMLTag(sFileContent, "target_directory", "./") ;
          log(" * target directory   : "+sTargetDir);

          //
          // connection string
          //
          sConnect = getXMLTag(sFileContent, "connection_string", null) ;
          log(" * connection string  : "+sConnect);
          
          //
          // auto compilation
          //
          st2 = getXMLTag(sFileContent, "compile_module", "false") ;
          if(st2.equalsIgnoreCase("true")) bCompile = true ;
          log(" * auto compilation   : "+bCompile);
          
          //
          // backup file extension
          //
          sBackExt = getXMLTag(sFileContent, "backup_ext", null) ;
          log(" * backup extension   : "+ (sBackExt != null ? sBackExt : "<null>"));
          
          //
          // new file extension
          //
          sNewExt = getXMLTag(sFileContent, "new_ext", null) ;
          log(" * new file extension : "+ (sNewExt != null ? sNewExt : "<null>"));
          
          //
          // comment line
          //
          sComment = getXMLTag(sFileContent, "comment", null) ;

          //
          // libraries case conversion
          //
          sLibCase = getXMLTag(sFileContent, "libraries_case", null) ;
          log(" * library conversion : "+sLibCase);
          
        }  

        /**------------------------------*
         *   properties for all modules  *
         *-------------------------------*/
        
        st = getXMLTag(sFileContent, "all_modules", null) ;
        if(st != null)
        {            
          // --------------------------------
          //  loop through all given modules
          // --------------------------------
          for(int z=0; z < module_list.size(); z++)
          {
            // open the module
            String sModule = (String)module_list.get(z);
            sCurrentModule = sModule ;
            if(!openModule(sModule)) continue ;

            //
            // make a backup copy
            //
            if(sBackupDir != null)
            {
              log("-- Backuping module to: " + sBackupDir + sFile + (sBackExt != null ? sBackExt : "") + "." + sExt) ;
              if(sBackExt == null)
               fm.save(sBackupDir + sFile + "." + sExt);  
              else
                fm.save(sBackupDir + sFile + sBackExt + "." + sExt);  
            }
            else
            {
              log("-- Backuping module to: " + sPath + sFile + (sBackExt != null ? sBackExt  : "") + "." + sExt);
              if(sBackExt == null)
                fm.save(sPath + sFile + "." + sExt);
              else
                fm.save(sPath + sFile + sBackExt + "." + sExt);
            }
              
            //
            // libraries case conversion
            //
            if(sLibCase != null) setLibraryCaseName(sLibCase);
            
            //
            // All canvases operations
            //
            String sCanvases = getXMLTag(sFileContent, "canvases", null) ;
            if(sCanvases != null)
            {
                // find all the module's canevases
                JdapiIterator cans = fm.getCanvases();
                //Loop through the canvases
                while(cans.hasNext())  
                { 
                  Canvas cv = (Canvas)cans.next();
                  if(! scanCanvases(sCanvases, cv.getName())) throw new ErrorException(" !! Process aborted !!");
                }   
            } 
            
            //
            // All blocks operations
            //
            String sBlocks = getXMLTag(sFileContent, "blocks", null) ;
            if(sBlocks != null)
            {
                // find all the module's blocks
                JdapiIterator blks = fm.getBlocks();
                //Loop through the blocks
                while(blks.hasNext())  
                { 
                  Block blk = (Block)blks.next();
                  if(! scanBlocks(sBlocks, blk.getName())) throw new ErrorException(" !! Process aborted !!");
                }   
            } 
  
            //
            // All items operations
            //
            String sItems = getXMLTag(sFileContent, "items", null) ;
            if(sItems != null)
            {
                List l = getXMLTagArray(sItems, "item") ;
                boolean b=true ;
                String sType   = "" ;
                String sName   = "" ;
                String sProp   = "" ;
                String sValue  = "" ;              
                String sErr    = "" ;                
                String sItm = null ;
                String sBlk = null ;
                int    iPos ;
                if(l.size()>0)
                {
                    for(j=0; j<l.size(); j++)
                    {
                        st = (String)l.get(j) ;
                        sType   = getXMLTag(st, "type", null) ;
                        sName   = getXMLTag(st, "name", null) ;
                        sProp   = getXMLTag(st, "property", null) ;
                        sValue  = getXMLTag(st, "value", null) ;              
                        sErr    = getXMLTag(st, "on_error", "continue") ;
                        boolean bProcess = false ;
                        if(sName != null) {
                            sItm   = getXMLTag(st, "name", null) ;
                            iPos   = sName.indexOf(".") ;
                            if(iPos > -1) {
                               sBlk = sName.substring(0,iPos); ;
                               sItm = sName.substring(iPos+1);
                            }
                            if(! doAction("ITEM", "UPDATE", sBlk, sItm, sProp, sValue, sErr)) throw new ErrorException(" !! Process aborted !!");
                        }
                        else
                        {
                            if(sType == null) sType = "all" ;                          
                            // find all the module's blocks
                            JdapiIterator blks = fm.getBlocks();
                            //Loop through the blocks
                            while(blks.hasNext())  
                            { 
                              Block blk = (Block)blks.next();
                              // find all the block's items
                              JdapiIterator its = blk.getItems();
                              //Loop through the items
                              while(its.hasNext())  
                              { 
                                bProcess = false ;
                                Item itm = (Item)its.next();
                                int iType = itm.getItemType();
                                // do we have to process this item ?
                                if     (sType.equalsIgnoreCase("all"))                       bProcess = true ;
                                else if(sType.equalsIgnoreCase("BeanArea")    && iType == 0) bProcess = true ;
                                else if(sType.equalsIgnoreCase("Chart")       && iType == 1) bProcess = true ;
                                else if(sType.equalsIgnoreCase("CheckBox")    && iType == 2) bProcess = true ;
                                else if(sType.equalsIgnoreCase("DisplayItem") && iType == 3) bProcess = true ;
                                else if(sType.equalsIgnoreCase("Tree")        && iType == 4) bProcess = true ;
                                else if(sType.equalsIgnoreCase("Image")       && iType == 5) bProcess = true ;
                                else if(sType.equalsIgnoreCase("List")        && iType == 6) {
                                   if(itm.getListStyle() < 2) bProcess = true ;
                                }   
                                else if(sType.equalsIgnoreCase("Combo")        && iType == 6) {
                                   if(itm.getListStyle() == 2) bProcess = true ;
                                }                                
                                else if(sType.equalsIgnoreCase("PushButton")  && iType == 7) bProcess = true ;
                                else if(sType.equalsIgnoreCase("RadioButton") && iType == 8) bProcess = true ;
                                else if(sType.equalsIgnoreCase("TextItem")    && iType == 9) bProcess = true ;
                                //System.out.println("item->"+itm.getName()+" - "+itm.getItemType()+ "  type="+sType+"  process:"+bProcess);                
                                if(bProcess)
                                {
                                  if(! doAction("ITEM", "UPDATE", blk.getName(), itm.getName(), sProp, sValue, sErr)) throw new ErrorException(" !! Process aborted !!");
                                }
                              }                   
                            }   
                        }
                    }
                }            
            } 

            //
            // All alerts operations
            //
            String sAlerts = getXMLTag(sFileContent, "alerts", null) ;
            if(sAlerts != null)
            {
                // find all the module's alerts
                JdapiIterator alts = fm.getAlerts();
                //Loop through the alerts
                while(alts.hasNext())  
                { 
                  Alert alt = (Alert)alts.next();
                  if(! scanAlerts(sAlerts, alt.getName())) throw new ErrorException(" !! Process aborted !!");
                }   
            }
                          
            //
            // All windows operations
            //
            String sWindows = getXMLTag(sFileContent, "windows", null) ;
            if(sWindows != null)
            {
                // find all the module's windows
                JdapiIterator wins = fm.getWindows();
                //Loop through the windows
                while(wins.hasNext())  
                { 
                  Window win = (Window)wins.next();
                  if(! scanWindows(sWindows, win.getName())) throw new ErrorException(" !! Process aborted !!");
                }   
            }
                      
          if(sNewExt != null)
              fm.setName(sFile + sNewExt);
          else
              fm.setName(sFile);
          if(sComment != null) fm.setComment(sComment.replaceFirst(":SYSDATE",dateNewFormat) + fm.getComment());
          if(sTargetDir != null) 
          {
            if(sNewExt != null)
              sFileNew = sTargetDir + sFile + sNewExt + "." + sExt ;
            else
              sFileNew = sTargetDir + sFile + "." + sExt ;
                
          }  
          else 
          {
            if(sNewExt != null)
              sFileNew = sPath + sFile + sNewExt + "." + sExt ;
            else
               sFileNew = sPath + sFile + "." + sExt ;
          }  
          log("-- Saving new module to: " + sFileNew);
          fm.save(sFileNew);
          
          // --------------
          // compilation ?
          // --------------
          if(bCompile && (!sConnect.equalsIgnoreCase("") && sConnect != null))
             {boolean b = compileModule(sFileNew);}
                  
        }
        //else throw new ErrorException("configuration file file.xml not found");
    
  
      }
      
      /**--------------------------------*
       *  properties for single modules  *
       *---------------------------------*/
      String singles = getXMLTag(sFileContent, "single_modules", null) ;
      if(singles != null)
      {
        List l = getXMLTagArray(singles, "module") ;
        boolean b=true ;
        if(l.size()>0)
        {
           for(j=0; j<l.size(); j++)
           {
              String sModuleContent = (String)l.get(j) ;            
              String sModuleName = getXMLTag(sModuleContent, "name", null) ;
              String sCompile = getXMLTag(sModuleContent, "compile", "false") ;
              if(sModuleName != null)
              {
                sCurrentModule = sModuleName ;
                if(!openModule(sModuleName)) continue ;
                    
                //
                // make a backup copy
                //
                if(sBackupDir != null)
                {
                  log("-- Backuping module to: " + sBackupDir + sFile + (sBackExt != null ? sBackExt  : "") + "." + sExt);
                  if(sBackExt == null)
                    fm.save(sBackupDir + sFile + "." + sExt);  
                  else
                     fm.save(sBackupDir + sFile + sBackExt + "." + sExt);  
                }
                else
                {
                  log("-- Backuping module to: " + sPath + sFile + (sBackExt != null ? sBackExt  : "") + "." + sExt);
                  if(sBackExt == null)
                    fm.save(sPath + sFile + "." + sExt);
                  else
                    fm.save(sPath + sFile + sBackExt + "." + sExt);
                }
                  
                // -------------------
                //  scan for triggers
                // -------------------
                if(! scanTriggers(sModuleContent)) throw new ErrorException(" !! Process aborted !!");
      
                // ------------------------
                //  scan for program units
                // ------------------------
                if(! scanProgramUnits(sModuleContent)) throw new ErrorException(" !! Process aborted !!");
      
                // ----------------------
                //  scan for libraries
                // ----------------------
                if(! scanLibraries(sModuleContent)) throw new ErrorException(" !! Process aborted !!");          
      
                // --------------------------------------
                //  scan for subclassed objects from olb
                // --------------------------------------
                if(! scanSubclassedObjects(sModuleContent)) throw new ErrorException(" !! Process aborted !!");          
      
                // ---------------------
                //  scan for canevases
                // ---------------------
                if(! scanCanvases(sModuleContent, null)) throw new ErrorException(" !! Process aborted !!");          
                  
                // ---------------------
                //  scan for windows
                // ---------------------
                if(! scanWindows(sModuleContent, null)) throw new ErrorException(" !! Process aborted !!");
                  
                // ---------------------
                //  scan for blocks
                // ---------------------
                if(! scanBlocks(sModuleContent, null)) throw new ErrorException(" !! Process aborted !!");
                  
                // ---------------------
                //  scan for items
                // ---------------------
                if(! scanItems(sModuleContent, null)) throw new ErrorException(" !! Process aborted !!");

                // ---------------------
                //  scan for alerts
                // ---------------------
                if(! scanAlerts(sModuleContent, null)) throw new ErrorException(" !! Process aborted !!");
                
                if(sNewExt != null)
                  fm.setName(sFile + sNewExt);
                else
                  fm.setName(sFile);
                if(sTargetDir != null) 
                {
                  if(sNewExt != null)
                    sFileNew = sTargetDir + sFile + sNewExt + "." + sExt ;
                  else
                    sFileNew = sTargetDir + sFile + "." + sExt ;
                }  
                else 
                {
                  if(sNewExt != null)
                    sFileNew = sPath + sFile + sNewExt + "." + sExt ;
                  else
                    sFileNew = sPath + sFile + "." + sExt ;
                }
                log("-- Saving new module to: " + sFileNew);
                fm.save(sFileNew);

                // --------------
                // compilation ?
                // --------------
                if(sCompile.equalsIgnoreCase("true") && (!sConnect.equalsIgnoreCase("") && sConnect != null))
                  b = compileModule(sFileNew);
              
              }
        
           }
        }   
            
      }
      
      
      //Frees the underlying native object. 
      if(fm != null) fm.destroy();  
    
    } // for(int z=0; z < module_list.size(); z++)
   } // if(st != null)    

    catch (ErrorException ee)
     {
       printError(ee.toString());
     }
     
    catch (JdapiException jex)
    {
      printError(jex.toString());
    }
   
    // Clean up
    Jdapi.shutdown();
    closeLogFile();
  }
  
  /***************************************************************************************/
  /**-------------------*
   *     M  A  I  N     *
   *--------------------*/
  // to test from JDev for instance
  public static void main(String[] args)
  {
    String sXMLFile = "c:/jdapi.xml" ;
    System.out.println("JDAPIUtil launched by main() procedure:");
    if(args.length > 0) sXMLFile = args[0];
    JDAPIUtil jd = new JDAPIUtil();
    jd.doProcess(sXMLFile);
  }
  /***************************************************************************************/
  
    /**---------------------------------------------*
     *  public procedure to switch logging on/off   *
     *----------------------------------------------*/
    public void setLog(boolean b)
    {
      bLog = b ;
    }
  
    /**---------------------------------------------*
     *  public procedure to switch logging on/off   *
     *----------------------------------------------*/
    public String getLogContent()
    {
      log("getLogContent()"+sLogFile);
      return readFile(sLogFile) ;
    }
  
    /**---------------------------------------------*
     *  public procedure to set the log file name   *
     *----------------------------------------------*/
    public void setLogFilename(String sFilename)
    {
      sLogFile = sFilename ;
    }
    
    /**-----------------------*
     *   manage all actions   *
     *------------------------*/
     boolean doAction(String s1, String s2, String s3, String s4, String s5, String s6, String s7)
    {
      if(s1.equalsIgnoreCase("TRIGGER"))
      {
         if(s2.equalsIgnoreCase("ADD"))              return addTrigger(s3, s5, s6, s7);
         else if(s2.equalsIgnoreCase("REPLACE"))     return updTrigger(s2,s3, s4, s5, s7, "");
         else if(s2.equalsIgnoreCase("HIERARCHY"))   return updTrigger(s2,s3, "", "", s7, s5);
         else if(s2.equalsIgnoreCase("DELETE_CODE")) return updTrigger(s2,s3, "", s5, s7, "");
         else if(s2.equalsIgnoreCase("REMOVE"))      return rmTrigger(s3, s7);        
      }
      else if(s1.equalsIgnoreCase("PROCEDURE")) 
      {
         if(s2.equalsIgnoreCase("ADD"))              return addProc(s3, s5, s6, s7);
         else if(s2.equalsIgnoreCase("REPLACE"))     return updProc(s2, s3, s4, s5, s7);
         else if(s2.equalsIgnoreCase("DELETE_CODE")) return updProc(s2, s3, "", s5, s7);
      }
      else if(s1.equalsIgnoreCase("LIBRARY")) 
      {
         if(s2.equalsIgnoreCase("ATTACH"))           return addLib(s3, s7);
         else if(s2.equalsIgnoreCase("DETACH"))      return rmLib(s3, s7);
      }
      else if(s1.equalsIgnoreCase("OLB")) 
      {
         return addOlb(s3, s4, s5, s6, s7);
      }
      else if(s1.equalsIgnoreCase("CANVAS")) 
      {
         if(s2.equalsIgnoreCase("UPDATE"))   return updCanvas(s3, s4, s5, s6);
      }      
      else if(s1.equalsIgnoreCase("WINDOW")) 
      {
         if(s2.equalsIgnoreCase("UPDATE"))   return updWindow(s3, s4, s5, s6);
      }
      else if(s1.equalsIgnoreCase("ALERT")) 
      {
         if(s2.equalsIgnoreCase("UPDATE"))   return updAlert(s3, s4, s5, s6);
      }
      else if(s1.equalsIgnoreCase("BLOCK")) 
      {
         //if(s2.equalsIgnoreCase("ADD"))             return addProc(s3, s5, s6, s7);
         if(s2.equalsIgnoreCase("UPDATE"))     return updBlock(s3, s4, s5, s6);
        // else if(s2.equalsIgnoreCase("REMOVE"))     return updProc(s2, s3, "", s5, s7);
      }      
      else if(s1.equalsIgnoreCase("ITEM")) 
      {
         if(s2.equalsIgnoreCase("ADD"))             return addProc(s3, s5, s6, s7);
         else if(s2.equalsIgnoreCase("UPDATE"))     return updItem(s3, s4, s5, s6, s7);
         else if(s2.equalsIgnoreCase("REMOVE"))     return updProc(s2, s3, "", s5, s7);
      }      
      return true ;
    }
    
    /**----------------*
     *  add a trigger  *
     *-----------------*/
    boolean addTrigger(String sName, String sText, String sPos, String sErr)
    {
         try
          {
            Trigger trg = Trigger.find(fm,sName);
            String tri_txt = trg.getTriggerText(); 
            if(sPos.equalsIgnoreCase("start")) 
               tri_txt = sText + "\n" + tri_txt ;
            else
               tri_txt = tri_txt + "\n" +  sText ;
            trg.setTriggerText (tri_txt);      
          }
          catch (Exception e) 
          {            
            // Create the trigger 
            try 
            { 
              Trigger trg = new Trigger(fm, sName ); 
              trg.setTriggerText(sText); 
            } 
            catch( Exception exc) 
            { 
                  System.setErr(System.out);
                  printError("addTrigger() Unable to create trigger: "+sName+". Error details: " + exc.toString());
                  //exc.printStackTrace(); 
                  if(sErr.equalsIgnoreCase("stop"))  return false ;
            }  
         }
         return true ;
    }
    

    /**-------------------*
     *  update a trigger  *
     *--------------------*/
    boolean updTrigger(String sAction, String sName, String sReplace, String sText,String sErr, String sHierarchy)
    {
         try
          {
            Trigger trg = Trigger.find(fm,sName);
            String txt = "" ;
            if((sAction.equalsIgnoreCase("REPLACE")) &&  ((sReplace != null) && (sText != null)))
            {
              //txt = trg.getTriggerText().replaceAll(sReplace, sText); 
              txt = _replaceAll(trg.getTriggerText(), sReplace, sText); 
              trg.setTriggerText (txt);      
            }
            if((sAction.equalsIgnoreCase("DELETE_CODE")) &&  (sText != null))
            {
              //log("updTrigger() deleting code:"+sText);
              //txt = trg.getTriggerText().replaceAll(sText, " "); 
              txt = _replaceAll(trg.getTriggerText(),sText, " "); 
              trg.setTriggerText (txt);      
            }            
            if(sHierarchy != null)
            {
              if(sHierarchy.equalsIgnoreCase("BEFORE"))        trg.setExecuteHierarchy(BEFORE);
              else if(sHierarchy.equalsIgnoreCase("AFTER"))    trg.setExecuteHierarchy(AFTER);
              else if(sHierarchy.equalsIgnoreCase("OVERRIDE")) trg.setExecuteHierarchy(OVERRIDE);
            }
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updTrigger() Unable to find trigger: "+sName+". Error details: " + e.toString());
              //e.printStackTrace(); 
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }


    /**-------------------*
     *  remove a trigger  *
     *--------------------*/
    boolean rmTrigger(String sName, String sErr)
    {
         try
          {
            Trigger trg = Trigger.find(fm,sName);
            trg.destroy();
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("rmTrigger() Unable to find trigger: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }    
    
    /**---------------------*
     *  add a program unit  *
     *----------------------*/
    boolean addProc(String sName, String sText, String sPos, String sErr)
    {
         try
          {
            ProgramUnit pu = ProgramUnit.find(fm,sName);
            String txt = pu.getProgramUnitText(); 
            if(sPos.equalsIgnoreCase("start")) 
               txt = sText + "\n" + txt ;
            else
               txt = txt + "\n" +  sText ;
            pu.setProgramUnitText(txt);
          }
          catch (Exception e) 
          {            
            // Create the program unit
            try 
            { 
              ProgramUnit pu = new ProgramUnit(fm,sName);
              pu.setProgramUnitText(sText);
            } 
            catch( Exception exc) 
            { 
                  System.setErr(System.out);
                  printError("addProc() Unable to create program unit: "+sName+". Error details: " + exc.toString());
                  //exc.printStackTrace(); 
                  if(sErr.equalsIgnoreCase("stop"))  return false ;
            }  
         }
         return true ;
    }
    

    /**------------------------*
     *  update a program unit  *
     *-------------------------*/
    boolean updProc(String sAction, String sName, String sReplace, String sText, String sErr)
    {
         try
          {
            ProgramUnit pu = ProgramUnit.find(fm,sName);
            String txt = pu.getProgramUnitText(); 
            if( sAction.equalsIgnoreCase("REPLACE") && ((sReplace != null) && (sText != null)))
            {
                //txt = pu.getProgramUnitText().replaceAll(sReplace, sText); 
                txt = _replaceAll(pu.getProgramUnitText(),sReplace, sText); 
                pu.setProgramUnitText(txt);
            }
            if( sAction.equalsIgnoreCase("DELETE_CODE") && (sText != null) )
            {
                //txt = pu.getProgramUnitText().replaceAll(sText, " "); 
                txt = _replaceAll(pu.getProgramUnitText(),sText, " "); 
                pu.setProgramUnitText(txt);
            }
          }
          catch (Exception exc) 
          {            
              System.setErr(System.out);
              printError("updProc() Unable to find program unit: "+sName+". Error details: " + exc.toString());
              //exc.printStackTrace(); 
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }

    /**-------------------*
     *  attach a library  *
     *--------------------*/
    boolean addLib(String sName, String sErr)
    {
         try
          {
            AttachedLibrary lib1 = new AttachedLibrary(fm,sName); 
            log("-- library attached: "+sName);
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("addLib() Unable to attach library: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }    

    /**--------------------*
     *  detach a library   *
     *---------------------*/
    boolean rmLib(String sName, String sErr)
    {
         try
          {
              AttachedLibrary lib = AttachedLibrary.find(fm,sName);
              lib.destroy();
              log("-- library removed: "+sName);
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("rmLib() Unable to dettach library: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    } 


    /**----------------------------------*
     *  add an Object library component  *
     *-----------------------------------*/
    boolean addOlb(String sName, String sTab, String sObject, String sLabel, String sErr)
    {
         int i = 1 ;
         if(sLabel.equalsIgnoreCase("") || sLabel== null)
         {
           printError("addOlb() ["+sName+"/"+sTab+"]: target name cannot be null");
           if(sErr.equalsIgnoreCase("stop"))  return false ;
           else return true ;
         }  
         try
          {
            int iPos = 0;//Integer.parseInt(sObject) ;
            JdapiObject obj1 = null;
            // subclass from object library 
            ObjectLibrary obl1=ObjectLibrary.open(sName); 
            i=2;
            ObjectLibraryTab tab1 =ObjectLibraryTab.find(obl1,sTab); 
            i=3;
            JdapiIterator it = tab1.getOwnedObjects();
            while(it.hasNext()) {
                iPos++;
                Object ob = it.next() ;
                if(ob instanceof oracle.forms.jdapi.ObjectGroup ) {
                    oracle.forms.jdapi.ObjectGroup og = (oracle.forms.jdapi.ObjectGroup)ob ;
                    obj1 = (JdapiObject)og ;                    
                    if(og.getName().equalsIgnoreCase(sLabel)) break ;
                }
            }            
            
            //JdapiObject obj1 = tab1.findObjectByPosition(iPos);
            obj1.createSubclassedChild( fm,sLabel); 
            log("-- addOlb() OLB component subclassed from:"+sName+"/"+sTab+"/"+sObject);            
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              String sLevel = "" ;
              switch(i)
              {
                case 1: sLevel = "[Object library not found:"+sName+"]"; break;
                case 2: sLevel = "[Tab not found:"+sTab+"]"; break;
                case 3: sLevel = "[Component not found:"+sObject+"]"; break;
                case 4: sLevel = "[Subclassing]"; break;
              }
              printError("addOlb() "+sLevel+": "+sName+"-"+sTab+"-"+sObject+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }    
    
    /**------------------*
     *   update canvas   *
     *-------------------*/
     boolean updCanvas(String sName, String sProperty, String sValue, String sErr) {
         //log("updCanvas "+sName+"-"+sProperty+":"+sValue);
         try
          {
            Canvas cv = Canvas.find(fm, sName);
            CanvasObject canvas = new CanvasObject();
            // set the canvas property
            boolean b = canvas.setProperty(cv,sProperty,sValue);
            if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updCanvas() Unable to find canvas: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
     }
    
    
    /**------------------*
     *   update alert   *
     *-------------------*/
     boolean updAlert(String sName, String sProperty, String sValue, String sErr) {
         //log("updAlert "+sName+"-"+sProperty+":"+sValue);
         try
          {
            Alert alt = Alert.find(fm, sName);
            AlertObject alert = new AlertObject();
            // set the alert property
            boolean b = alert.setProperty(alt,sProperty,sValue);
            if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updAlert() Unable to find alert: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
     }
    
    /**------------------*
     *   update window   *
     *-------------------*/
     boolean updWindow(String sName, String sProperty, String sValue, String sErr) {
         //log("updWindow "+sName+"-"+sProperty+":"+sValue);
         try
          {
            Window win = Window.find(fm, sName);
            WindowObject window = new WindowObject();
            // set the window property
            boolean b = window.setProperty(win,sProperty,sValue);
            if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updWindow() Unable to find window: "+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
     }


    /**-------------------*
     *   update a block   *
     *--------------------*/
    boolean updBlock(String sBlock, String sProperty, String sValue, String sErr)
    {
         try
          {
            Block blk = Block.find(fm, sBlock);
            BlockObject block = new BlockObject();
            // set the block property
            boolean b = block.setProperty(blk,sProperty,sValue);
            if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updBlock() Unable to find block: "+sBlock+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }


    /**-------------------*
     *   update an item   *
     *--------------------*/
    boolean updItem(String sBlock, String sName, String sProperty, String sValue, String sErr)
    {
         try
          {
            // multi-block items ?
            if(sBlock == null) {
                // find all the module's blocks
                JdapiIterator blks = fm.getBlocks();
                //Loop through the blocks
                while(blks.hasNext())  
                { 
                  Block blk = (Block)blks.next();
                  try{
                    Item itm = Item.find(blk,sName);
                    ItemObject io = new ItemObject();
                    // set the item property
                    boolean b = io.setProperty(itm,sProperty,sValue);
                    if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
                  }
                  catch(Exception ex1){;}
                }                   
            }
            else
            {
                Block blk = Block.find(fm, sBlock);
                Item itm = Item.find(blk,sName);
                ItemObject io = new ItemObject();
                // set the item property
                boolean b = io.setProperty(itm,sProperty,sValue);
                if(!b && sErr.equalsIgnoreCase("stop"))  return false ;
            }
          }
          catch (Exception e) 
          {            
              System.setErr(System.out);
              printError("updItem() Unable to find item: "+sBlock+"."+sName+". Error details: " + e.toString());
              if(sErr.equalsIgnoreCase("stop"))  return false ;
         }
         return true ;
    }



     /**------------------------------*
      *    read a text from a file    *
      *-------------------------------*/
     private String readFile(String sFileName)
      {
        BufferedReader fileIn=null;
        StringBuffer sb = new StringBuffer();
        String line = "" ;
        bFileOk = false ;
       
        // Open the file
        try
        {
          fileIn = new BufferedReader( new FileReader(sFileName));
        } catch (FileNotFoundException e)
        {
          System.out.println("readFile() open error: "+sFileName+" not found ");
          JOptionPane.showMessageDialog(null,"File: "+sFileName+" not found","JDAPIUtil - XML file",JOptionPane.ERROR_MESSAGE);
          return "readFile() open error: "+sFileName+" not found " ;
        }
          catch (IOException e)
        {
          System.out.println("readFile() open error: "+e);
          JOptionPane.showMessageDialog(null,"Open error: "+sFileName,"JDAPIUtil - XML file",JOptionPane.ERROR_MESSAGE);
          return "readFile() open error: "+e;
        }     
        // read the file
        try
        {
          while((line=fileIn.readLine())!= null)
          {
              sb.append(line + '\n');
          }
        }catch (Exception e)
        {
          System.out.println("readFile() reading error:"+e);
          return "readFile() reading error:"+e ;
        }
        try
        {
          fileIn.close();
        } catch (Exception e)
        {
          System.out.println("readFile() closing error:"+e);
          return "readFile() closing error:"+e ;
        }
        bFileOk = true ;
        return sb.toString(); 
      }


    /**------------------------------------------*
     *   returns an Array list of found tokens   *
     *-------------------------------------------*/
    List getXMLTagArray(String sXMLContent, String sTag)
    {
      String       sLigne      = "" ;
      String       sXml        = "" ;
      String       sTagStart   = "<" + sTag + ">" ;
      String       sTagEnd     = "</" + sTag + ">" ;
      int          iIndex      = 0;
      int          iTagDeb     = 0 ;
      int          iTagEnd     = 0 ;
      int          iTagLength  = 0 ;
      int          iLigLength  = 0 ;      
      boolean      bCont       = true ;
      
      List liste = new ArrayList(5) ; 
      
      if( sXMLContent != null && sTag != null )
      {
        sXml = sXMLContent ;
        iTagLength = sTagStart.length() ;
        while(bCont)
        {
          iIndex++ ;
          iTagEnd = sXml.indexOf(sTagEnd) ;
          iTagDeb = sXml.indexOf(sTagStart) ;        
          iLigLength = (iTagEnd - ( iTagDeb + iTagLength ) ) ;
          if(iTagDeb > -1 && iTagEnd > -1 )
          {
             //log("deb="+iTagDeb+"  end="+iTagEnd+" len="+iLigLength);
             sLigne = sXml.substring(iTagDeb+iTagLength, iTagDeb+iTagLength+iLigLength);
             liste.add((Object)sLigne.trim());
             //log("lig=["+sLigne.trim()+"]");
             int i = iTagEnd + sTagEnd.length() ;
             if(i >=sXml.length()) bCont = false ;
             else sXml = sXml.substring(i);
             //log("rest=["+sXml+"]");
          }
          else bCont = false ;
        }
      }
      return liste ;
    }
    
    /**----------------------------------*
     *    returns the content of a tag   *
     *-----------------------------------*/
    String getXMLTag(String sXMLContent, String sTag, String sNullValue)
    {
      String       sLigne      = null ;
      String       sXml        = "" ;
      String       sTagStart   = "<" + sTag + ">" ;
      String       sTagEnd     = "</" + sTag + ">" ;
      int          iTagDeb     = 0 ;
      int          iTagEnd     = 0 ;
      int          iTagLength  = 0 ;
      int          iLigLength  = 0 ;      
      
      if( sXMLContent != null && sTag != null )
      {
        sXml = sXMLContent ;
        iTagLength = sTagStart.length() ;
        iTagEnd = sXml.indexOf(sTagEnd) ;
        iTagDeb = sXml.indexOf(sTagStart) ;        
        iLigLength = (iTagEnd - ( iTagDeb + iTagLength ) ) ;
        if(iTagDeb > -1 && iTagEnd > -1 )
        {
           //log("deb="+iTagDeb+"  end="+iTagEnd+" len="+iLigLength);
           sLigne = sXml.substring(iTagDeb+iTagLength, iTagDeb+iTagLength+iLigLength);
           sLigne = sLigne.trim();
        }
      }
      if(sLigne != null && sLigne.equalsIgnoreCase("")) sLigne=null ;
      if((sLigne == null) && (sNullValue != null)) sLigne = sNullValue ;

      return sLigne ;
    }

    /**-----------------------------------------*
     *  set the new attached library case name  *
     *------------------------------------------*/
    void setLibraryCaseName(String sCase) {
       //log("-- Lib -------------------");
       //log(" setLibraryCaseName():"+sCase);
       String sName = "" ;
       int itot=0;
       // find all the module's libraries
       JdapiIterator libs = fm.getAttachedLibraries();
       libs.goStart();
       //Loop through the attached libraries
        while (libs.hasNext())   
        {  
          AttachedLibrary lib = (AttachedLibrary)libs.next();   
          itot++;  
        }        
        log(" Total libraries found:"+ itot);
        //Reloading the attached libraries  
        libs = fm.getAttachedLibraries();   
        libs.goStart();  
        String libraries[] = new String[itot+5];  
        //try   
        //{   
        int i=0;  
        // loop to fill the attached libraries names into an array of String (libraries)  
        while (libs.hasNext())  
        {  
            
          AttachedLibrary lib = (AttachedLibrary)libs.next();   
          // Identify the library name.   
          libraries[i] = lib.getName();   
          log(" -- finding lib: "+lib.getName());  
          i++;  
        }  
        
        // loop in the array to detach all libraries and reattach again in lower case  
        for (int j=0;j<itot;j++)  
        {  
          AttachedLibrary lib, lib1;  
          lib=AttachedLibrary.find(fm,libraries[j]);  
          lib.destroy();  
          log(" -- deattaching lib: "+libraries[j]);  
          if(sCase.equalsIgnoreCase("to_lower_case")) lib1 = new AttachedLibrary(fm,libraries[j].toLowerCase());   
          else if(sCase.equalsIgnoreCase("to_upper_case")) lib1 = new AttachedLibrary(fm,libraries[j].toUpperCase());   
          log(" -- reattaching lib: "+libraries[j].toLowerCase());  
        }          
       /* 
       while (libs.hasNext())  
       { 
         AttachedLibrary lib = (AttachedLibrary)libs.next(); 
         if(lib != null)
         if(sCase.equalsIgnoreCase("to_lower_case")) sName = lib.getName().toLowerCase() ;
         else if(sCase.equalsIgnoreCase("to_upper_case")) sName = lib.getName().toUpperCase() ;
         log("-- library found:"+lib.getName());
         lib.destroy();
         AttachedLibrary lib2 = new AttachedLibrary(fm,sName);   
         log("-- library re-attached:"+sName);
       } 
       */
       //log(" Total libraries found:"+ i);
       //log("------------------- Lib --");        
    }

    // -------------------
    //  scan for triggers
    // -------------------
    boolean scanTriggers(String sSource)
    {
        List l = getXMLTagArray(sSource, "trigger") ;
        boolean b=true ;
        if(l.size()>0)
        {
          for(int j=0; j<l.size(); j++)
          {
            String st = (String)l.get(j) ;
            //log("TRG-->["+st2+"]");
            String sAction = getXMLTag(st, "action", null) ;
            String sName   = getXMLTag(st, "name", null) ;
            String sRepl   = getXMLTag(st, "replace", null) ;
            String sText   = getXMLTag(st, "text", null) ;
            String sPos    = getXMLTag(st, "position", "start") ;
            String sErr    = getXMLTag(st, "on_error", "continue") ;
            //log("-->"+sAction+"-"+sName+"- replace["+sRepl+"] by ["+sText+"]");
            b = doAction("TRIGGER", sAction, sName, sRepl, sText, sPos, sErr) ;
          }
        }
        return b ;
    }
    
    // ------------------------
    //  scan for program units
    // ------------------------
    boolean scanProgramUnits(String sSource)
    {
        List l = getXMLTagArray(sSource, "program_unit") ;
        boolean b=true ;
        if(l.size()>0)
        {
          for(int j=0; j<l.size(); j++)
          {
            String st = (String)l.get(j) ;
            //log("PU-->["+st2+"]");
            String sAction = getXMLTag(st, "action", null) ;
            String sName   = getXMLTag(st, "name", null) ;
            String sRepl   = getXMLTag(st, "replace", null) ;
            String sText   = getXMLTag(st, "text", null) ;
            String sPos    = getXMLTag(st, "position", "start") ;
            String sErr    = getXMLTag(st, "on_error", "continue") ;
            //log("-->"+sAction+"-"+sName+"-"+sText);
            b = doAction("PROCEDURE", sAction, sName, sRepl, sText, sPos, sErr) ;
          }
        }            
        return b ;
    }
    
    // ------------------------------
    //  scan for attached libraries
    // ------------------------------
    boolean scanLibraries(String sSource)
    {
        List l = getXMLTagArray(sSource, "library") ;
        boolean b=true ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
              String st = (String)l.get(j) ;
              String sAction = getXMLTag(st, "action", null) ;
              String sName   = getXMLTag(st, "name", null) ;
              String sErr    = getXMLTag(st, "on_error", "continue") ;
              //log("-->"+sAction+"-"+sName+"-"+sText);
              b = doAction("LIBRARY", sAction, sName, "", "", "", sErr) ;
            }
        }            
        return b ;
    }
    
    // ------------------------------
    //  scan for subclassed objects
    // ------------------------------
    boolean scanSubclassedObjects(String sSource)
    {
        List l = getXMLTagArray(sSource, "subclass_olb") ;
        boolean b=true ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                String sName   = getXMLTag(st, "lib_name", null) ;
                String sTab    = getXMLTag(st, "tab_name", null) ;
                String sObject = getXMLTag(st, "object_num", null) ;              
                String sLabel  = getXMLTag(st, "target_name", null) ;
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("OLB", "", sName, sTab, sObject, sLabel, sErr) ;
            }
        }            
        return b ;
    }
    
    // --------------------
    //   scan for blocks
    // --------------------
    boolean scanBlocks(String sSource,String sBlockName)
    {
        List l = getXMLTagArray(sSource, "block") ;
        boolean b=true ;
        String sName = sBlockName ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                if(sName == null) sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("BLOCK", "update", sName, sProp, sValue, sErr, "") ;
            }
        }            
        return b ;
    }
    
    // --------------------
    //   scan for items
    // --------------------
    boolean scanItems(String sSource,String sItemName)
    {
        List l = getXMLTagArray(sSource, "item") ;
        boolean b=true ;
        String sName = sItemName ;
        String sBlk = null ;
        String sItm = null ;
        int    iPos ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                if(sName == null) 
                {
                  sItm   = getXMLTag(st, "name", null) ;
                  iPos   = sName.indexOf(".") ;
                  if(iPos > -1) {
                     sBlk = sName.substring(0,iPos); ;
                     sItm = sName.substring(iPos+1);
                  }
                }  
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("ITEM", "update", sBlk, sName, sProp, sValue, sErr) ;
            }
        }            
        return b ;
    }

    // --------------------
    //  scan for canvases
    // --------------------
    boolean scanCanvases(String sSource,String sCanvasName)
    {
        List l = getXMLTagArray(sSource, "canvas") ;
        boolean b=true ;
        String sName = sCanvasName ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                if(sName == null) sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("scanCanvases-->"+sName+"-"+sProp+"-"+sValue);
                b = doAction("CANVAS", "update", sName, sProp, sValue, sErr, "") ;
            }
        }            
        return b ;
    }
    
    // --------------------
    //  scan for windows
    // --------------------
    boolean scanWindows(String sSource,String sWindowName)
    {
        List l = getXMLTagArray(sSource, "window") ;
        boolean b=true ;
        String sName = sWindowName ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                if(sName == null) sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("WINDOW", "update", sName, sProp, sValue, sErr, "") ;
            }
        }            
        return b ;
    }
    
    // --------------------
    //  scan for alerts
    // --------------------
    boolean scanAlerts(String sSource,String sAlertName)
    {
        List l = getXMLTagArray(sSource, "alert") ;
        boolean b=true ;
        String sName = sAlertName ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                if(sName == null) sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("ALERT", "update", sName, sProp, sValue, sErr, "") ;
            }
        }            
        return b ;
    }

    
    // -----------------
    //  scan for items
    // -----------------
    boolean scanItems(String sSource,String sBlockName, String sItemName)
    {
        List l = getXMLTagArray(sSource, "item") ;
        boolean b=true ;
        if(l.size()>0)
        {
            for(int j=0; j<l.size(); j++)
            {
                String st = (String)l.get(j) ;
                String sName   = getXMLTag(st, "name", null) ;
                String sProp   = getXMLTag(st, "property", null) ;
                String sValue  = getXMLTag(st, "value", null) ;              
                String sErr    = getXMLTag(st, "on_error", "continue") ;
                //log("-->"+sAction+"-"+sName+"-"+sText);
                b = doAction("WINDOW", "update", sName, "", sProp, sValue, sErr) ;
            }
        }            
        return b ;
    }

   /**------------------------*
    *   open a Forms module   *
    *-------------------------*/
   boolean openModule(String sModuleName)
   {
        String s = "-- proceeding module: " + sModuleName + " --" ;
        log(" ");
        log(pad('-',s.length()));
        log(s);
        log(pad('-',s.length()));

        f = new File(sModuleName);
        // path
        sPath = f.getParent()+f.separator;
        // file name
        sFile = f.getName();
        // file extension
        ie = sFile.indexOf(".") ;
        sExt = sFile.substring(ie+1);
        sFile = sFile.substring(0,ie);
        // extension must be .fmb
        if( ! sExt.equalsIgnoreCase("FMB"))
        {
          System.setErr(System.out);
          printError("Module : "+sModuleName+". File extension must be .FMB");
          return false ;
        }

        try{
          mod = JdapiModule.openModule(sModuleName);             
          fm = (FormModule)mod.getModule();
        }
        catch(Exception ex)
        {
          System.setErr(System.out);
          printError("Module : "+sModuleName+". Unable to open. Error details: " + ex.toString());
          return false ;
        }
        return true ;
    }
    
    /**---------------------*
     *   compile a module   *
     *----------------------*/
    boolean compileModule(String sFile)
    {
        log("-- Compilation of module:"+sFile);
        try {
          Jdapi.connectToDatabase(sConnect); 
        }
        catch (JdapiException e) 
        {
          System.setErr(System.out);
          printError("Database connection : Unable to connect to the specified database. The form has been modified but not compiled. Please compile it manually. Error details: " + e.toString());
          return false ;
        }
        // Compile the form to create a fmx. 
        if (Jdapi.isConnectedToDatabase() ) {
          try{
            fm.compile(); 
            log("   module: "+sFile+ " compiled.");
          }
          catch (JdapiException e) 
          {
            System.setErr(System.out);
            printError("Compilation : Unable to compile The form. Error details: " + e.toString());
            return false ;
          }
        }
        return true ;
    }

    /**-----------------------------------*
     *  replaceAll() kind 1.3 equivalent  *
     *------------------------------------*/
    String _replaceAll(String source, String pattern, String replace)
    {
        if (source!=null)
        {
        final int len = pattern.length();
        StringBuffer sb = new StringBuffer();
        int found = -1;
        int start = 0; 
        while( (found = source.indexOf(pattern, start) ) != -1) {
            sb.append(source.substring(start, found));
            sb.append(replace);
            start = found + len;
        }

        sb.append(source.substring(start));

        return sb.toString();
        }
        else return "";
    }


    public static boolean getBooleanValue(String sValue)
    {
      if(sValue.equalsIgnoreCase("true"))  return true ;
      else return false ;
    }
    
    /**----------------------------*
     *  Ouput messages on console  *
     *-----------------------------*/
    private void log( String msg)
    {
       if(bLog) System.out.println(msg) ;
       if(bLogOpen) writeLogFile(msg);
    }
    

    /**------------------*
     *   open log file   *
     *-------------------*/
    boolean openLogFile(String sFileName)
    {
      // open the file
      log("Opening log file: " + sFileName ) ;
      try
      {
        fileOut = new PrintWriter(new FileWriter(sFileName));
        bLogOpen = true ;
      } catch (Exception e)
      {
        printError("openLogFile opening error:"+e); 
        bLogOpen = false ;
        return false ;
      }  
      return true ;
    }
    
    /**-------------------*
     *   close log file   *
     *--------------------*/
    void closeLogFile()
    {
       try
       {
         if(fileOut != null) fileOut.close();
       } catch (Exception e)
       {
         printError("closeLogFile closing error:"+e);
       }       
       finally{bLogOpen = false ;}
    }
    
  /**----------------------*
   *   Write to log file   *
   *-----------------------*/
  private void writeLogFile(String sLine)
  {
      // write to the file
      try
        {
          fileOut.println( sLine ) ;
        } catch (Exception e)
        {
          printError(" writeLogFile writing error:"+e); 
        }
  }
        
   /**------------------------------------------------*
    *  Simply prints an error message to the console  *
    *-------------------------------------------------*/
   private void printError(String errorString)
   {
     System.out.println("  *** " + sCurrentModule + " ERROR *** " + errorString);
   }

    private String pad(char c1, int i1)
    {
      StringBuffer sb = new StringBuffer();
      for(int i=0 ; i < i1; i++) sb.append(c1);
      return sb.toString() ;
    }
    private String rpad(String s1, char c1, int i1)
    {
      StringBuffer sb = new StringBuffer(s1);
      for(int i=0 ; i < i1; i++) sb.append(c1);
      return sb.toString() ;
    }    
    private String lpad(String s1, char c1, int i1)
    {
      StringBuffer sb = new StringBuffer();
      for(int i=0 ; i < i1; i++) sb.append(c1);
      sb.append(s1);
      return sb.toString() ;
    }        
    
} 
 

